<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
/** Defines import/export capabilities for the library.
 * The calling controller should have access to all models (i.e. have them
 * in its $uses variable).
 */
class ImportExportComponent extends Object {
	public $components = array('DemyoConfiguration', 'File', 'SpreadsheetImportExport');
	private $controller;
	private $errors;
	
	public function startup(&$controller) {
		$this->controller =& $controller;
		$this->errors = array();		
	}
	
	/** Logs an error message.
	 *
	 * @param string $message The message to log. Will be localized.
	 * @return boolean false (always)
	 */
	private function error($message) {
		if (file_exists(IMPORT_STATUS_FILE))
			unlink(IMPORT_STATUS_FILE); // Clean
		$this->errors[] =$message;
		return false;
	}
	
	/** Returns a log of errors that occured during processing.
	 *
	 * @return string
	 */
	public function getErrors() {
		return $this->errors;
	}
	
	/** Exports the database.
	 *
	 * @param string $method The export type.
	 * @param boolean $withResources true if associated resources (most notably
	 * images) should be packed with the data.
	 * @return An array with mimetype and content or zip file info, depending
	 * on wether the resources have been included or not. False in case of
	 * error, see {@link #getErrors} for the log.
	 */
	public function export($method, $withResources = false, $onlyWishlist = false) {
		// Just in case we would need it, act before $method is modified
		$retZipName = "demyo-export-$method.zip";
		
		$exported = array();
		$options = array();
		
		// Detect method
		switch ($method) {
			case 'xml+xsl':
				$options['XSL'] = true;
				// Fall-through
			case 'xml':
				$exported['mimetype'] = 'application/xml';
				$method = 'to_xml';
				break;
			case 'csv':
				$method = 'to_csv';
				$exported['mimetype'] = 'text/csv';
				break;
			case 'xls':
				$method = 'to_xls';
				$exported['mimetype'] = 'application/vnd.ms-excel';
				break;
			case 'xlsx':
				$method = 'to_xlsx';
				$exported['mimetype'] = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet';
				break;
			default:
				return $this->error(__('Invalid export method', true));
		}
		$options['wishlist'] = $onlyWishlist;
		
		// Run export
		if (! $ret = $this->$method($options))
			return $this->error(__s('Failed to export with method %s', $method));
		list($filename, $content) = $ret;
		
		// Return data, or build a zip with the resources
		if (! $withResources)
			return array_merge($exported, array('content' => $content, 'filename' => $filename));
		
		// Reset the mimetype: we'll be sending a zip no matter what
		$exported['mimetype'] = 'application/zip';

		$zipName = tempnam(TMP, 'demyo-export-');
		unlink($zipName); // ZipArchive needs a .zip 
		$zip = new ExtendedZipArchive();
		$zipName .= '.zip';
		if ($zip->open($zipName, ZipArchive::CREATE) !== true)
			return $this->error(__('Failed to create ZIP archive', true));;
		
		$zip->addFromString($filename, $content);
		$zip->addDirectory(COLLECTION_IMAGES);
		
		$zip->close();
		//register_shutdown_function('unlink', $zipName); // Delete on exit, done from the view (meh, Cake doesn't look adapted for readfile transfers)
		
		Configure::write('Demyo.lastExport', time());
		$this->DemyoConfiguration->saveToFile();
		
		return array_merge($exported, array('zipFileName' => $retZipName, 'zipPath' => $zipName, 'filename' => $retZipName));
	}
	
	public function import($method, $formFileData, $forceImport = false) {
		if (file_exists(IMPORT_STATUS_FILE))
			return $this->error(__('Import already in progress', true));
		else
			touch(IMPORT_STATUS_FILE);
		
		switch ($method) {
			case 'xml':
				$method = 'from_xml';
				break;
			case 'csv':
				$method = 'from_csv';
				break;
			default:
				return $this->error(__('No such import method', true));
		}

		if ($formFileData['error'] !== UPLOAD_ERR_OK)
			return $this->error(__s('Upload failed with <a href="http://php.net/manual/en/features.file-upload.errors.php">error code %s</a>', $formFileData['error']));

		$ext = pathinfo($formFileData['name'], PATHINFO_EXTENSION);
		
		$parseable = $formFileData['tmp_name'];
		
		if ($ext === 'zip') { // Unzip
			
			// Make temp folder
			$tmpFolder = TMP.DS.'zip_tmp';
			if (file_exists($tmpFolder) && !$this->File->rm($tmpFolder))
				return $this->error(__('Failed to remove old temporary resource directory', true));
			if (! mkdir($tmpFolder, 0700))
				return $this->error(__('Failed to make temporary resource directory', true));
				
			// Unzip
			$zip = new ZipArchive();
			if ($zip->open($formFileData['tmp_name']) !== true)
				return $this->error(__('Failed to open ZIP archive', true));
			if (! $zip->extractTo($tmpFolder))
				$this->error(__('Failed to extract ZIP archive', true));
			$zip->close();
			
			// Set import file
			switch ($method) {
				case 'from_xml':
					$parseable = $tmpFolder.DS.'demyo.xml';
					break;
				default:
					return $this->error(__('Could not find the collection data in the archive', true));
			}
		}
		
		try {
			$ret =  $this->$method($parseable, $forceImport);
		} catch (SchemaMismatchException $e) {
			unlink(IMPORT_STATUS_FILE);
			throw $e;
		}
		
		if ($ret === true && $ext == 'zip') { // Restore images
			
			// Delete all files but root
			if (! $this->File->rm(COLLECTION_IMAGES, true))
				return $this->error(__('Failed to delete previous images', true));
			
			// Move contents from temporary extraction to final directory
			if (! $this->File->moveDirContent($tmpFolder.DS.COLLECTION_IMAGES_DIRNAME, COLLECTION_IMAGES))
				return $this->error(__('Failed to restore images', true));
			// Give rights in final directory. See bug #91
			if (! $this->File->chmod(COLLECTION_IMAGES, 0666))
				return $this->error(__('Failed to set rights for images', true));
		}
		
		unlink(IMPORT_STATUS_FILE);
		if (isset($tmpFolder))
			$this->File->rm($tmpFolder);
		return $ret;
	}
	
	public function getImportStatus() {
		if (is_readable(IMPORT_STATUS_FILE))
			return file_get_contents(IMPORT_STATUS_FILE);
		else
			return 'null';
	}
	
	private function to_xml($options) {
		$withXSL = isset($options['XSL']) && $options['XSL'];
		
		// Prepare header before trying to export (fast-fail)
		$ret = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
		$style = '';
		if ($withXSL) {
			$ret .= <<<XSL
<?xml-stylesheet href="#style" type="text/xsl"?>
<!DOCTYPE doc [
<!ATTLIST xsl:stylesheet id ID #REQUIRED>
]>

XSL;
			$style = file_get_contents(WWW_ROOT.DS.'files'.DS.'demyo.xsl');
			if ($style === false)
				return $this->error(__('Failed to open XSL stylesheet', true));
		}
		
		// We want a float with . as a decimal separator. number_format is not
		// suited for the job since it needs the number of decimals, which we don't know
		$demyoVersion = str_replace(',', '.', (string)Configure::read('Demyo.versionAsFloat'));
		
		// Build XML
		$output = new SimpleXMLElement('<?xml version="1.0" encoding="UTF-8"?><library demyo-version="'
			.$demyoVersion.'" schema-version="'.Configure::read('Demyo.schemaVersion').'">'.$style.'</library>');
		$totalEntries = 0;
		foreach (array('Publisher', 'Collection', 'Binding', 'Author', 'Tag',
			'Image', 'Series', 'Album', 'DerivativeType', 'Source', 'Derivative',
			'Borrower', 'AlbumsBorrower', 'AlbumPrice', 'DerivativePrice')
			as $modelName) {
		
			$totalEntries += $this->controller->$modelName->addToElement($output);
		}
		$output->addAttribute("entries", $totalEntries);
		
		// Format XML
		$doc = new DOMDocument('1.0', 'utf-8');
		$doc->formatOutput = true;
		$domnode = dom_import_simplexml($output);
		$domnode = $doc->importNode($domnode, true);
		$domnode = $doc->appendChild($domnode);
		
		return array('demyo.xml', $ret.$doc->saveXML($domnode));
	}
	
	private function to_csv($options) {
		return $this->SpreadsheetImportExport->exportToCSV($this->controller, $options);
	}
	
	private function to_xls($options) {
		return $this->SpreadsheetImportExport->exportToXLS($this->controller, $options);
	}

	private function to_xlsx($options) {
		return $this->SpreadsheetImportExport->exportToXLSX($this->controller, $options);
	}
	
	private function from_xml($parseable, $forceImport = false) {
		if (($xml = simplexml_load_file($parseable)) === false)
			return $this->error(__('Failed to load XML file', true));
		$xml = current($xml->xpath('//library'));
		
		$attributes = $xml->attributes();
		$entries = $attributes['entries'];
		
		if (!$forceImport
				&& (empty($attributes['schema-version'])
					|| $attributes['schema-version'] < Configure::read('Demyo.schemaVersion'))) {
			throw new SchemaMismatchException();
		}
		
		$reporter = new ProgressReporter($entries, IMPORT_STATUS_FILE);
		$reporter->setAutoWrite(true);
		$processed = 0;
		
		// When importing, the order does matter: the database potentially
		// checks references. Here is the list of all models, dependencies first.
		$models = array('Image', 'Publisher', 'Collection', 'Binding', 'Author', 'Tag',
			'Series', 'Album', 'DerivativeType', 'Source',
			'Derivative', 'Borrower', 'AlbumsBorrower', 'AlbumPrice',
			'DerivativePrice');

		// Delete in reverse order
		foreach(array_reverse($models) as $modelName) {
			$this->controller->$modelName->clear();
		}

		// Insert in regular order
		foreach (array('Image', 'Publisher', 'Collection', 'Binding', 'Author', 'Tag',
				'Series', 'Album', 'DerivativeType', 'Source',
				'Derivative', 'Borrower', 'AlbumsBorrower', 'AlbumPrice',
				'DerivativePrice')
				as $modelName) {
			
			if (!$this->controller->$modelName->importXML($xml, $reporter))
				return false;
		}
		
		return true;
	}
	
	private function from_csv($parseable, $forceImport = false) {
		$reporter = new ProgressReporter(-1, IMPORT_STATUS_FILE);
		$reporter->setAutoWrite(true);
		
		$warnings = $this->CSVImportExport->importFromCSV($this->controller, $parseable, $reporter);
		if (count($warnings) <= 0)
			return true;
		$this->errors = array_merge($this->errors, $warnings);
		return false;
	}
}

class ProgressReporter {
	/** Number of handled entries.
	 * @var int */
	private $current;
	
	/** Total entries to handle.
	 * @var int */
	private $total;
	
	/** Parts progress
	 * @var array
	 */
	private $parts;
	private $currentPart;
	private $outFile;
	private $autoWrite;
	
	function __construct($total, $outFile) {
		$this->current = 0;
		$this->total = $total;
		$this->parts = array();
		$this->currentPart = -1;
		$this->outFile = $outFile;
		$this->autoWrite = false;
	}
	
	function setAutoWrite($auto) {
		$this->autoWrite = $auto;
		if ($auto)
			$this->writeJSON();
	}
	
	function setCurrent($current) {
		$this->current = $current;
		if ($this->autoWrite)
			$this->writeJSON();
	}
	
	function createPart($name, $total) {
		$cur = ++$this->currentPart;
		$this->parts[$cur] = array('name' => $name, 'current' => 0, 'total' => $total);
		if ($this->autoWrite)
			$this->writeJSON();
	}
	
	/** Sets the current amount for the part.
	 * 
	 * @param $current int
	 * @param $updateOverallCurrent bool True if the overall current count
	 * should be adapted (newOverall = oldOverall + (newPart - oldPart))
	 */
	function setPartCurrent($current, $updateOverallCurrent = true) {
		if ($updateOverallCurrent)
			$this->current = $this->current - $this->parts[$this->currentPart]['current'] + $current;
		
		$this->parts[$this->currentPart]['current'] = $current;
		
		if ($this->autoWrite)
			$this->writeJSON();
	}
	
	function getPartCurrent() {
		return $this->parts[$this->currentPart]['current'];
	}
	
	function toJSON() {
		if (($cur = $this->currentPart) > -1) {
			$json = <<<HEREDOC
{
	current: {$this->current},
	total: {$this->total},
	part: "{$this->parts[$cur]['name']}",
	partCurrent: {$this->parts[$cur]['current']},
	partTotal: {$this->parts[$cur]['total']}
}
HEREDOC;
		} else {
			$json = <<<HEREDOC
{
	current: {$this->current},
	total: {$this->total},
	part: null,
	partCurrent: 0,
	partTotal: -1
}
HEREDOC;
		}
		return $json;
	}
	
	function writeJSON() {
		file_put_contents($this->outFile, $this->toJSON());
	}
}

/** Denotes a mismatch between the version of the schema of an import source
 * and the one of the current application.
 */
class SchemaMismatchException extends Exception {}

class ExtendedZipArchive extends ZipArchive {
	const REOPEN_EVERY = 200;
	
	private $openedFileName = null;
	private $activeHandles;
	
	/**
	 * Open a ZIP file archive
	 * @param filename string The file name of the ZIP archive to open.
	 * @param flags int[optional] The mode to use to open the archive.
	 */
	public function open($filename, $flags = null) {
		$ret = parent::open($filename, $flags);
		
		if ($ret === true) {
			$this->activeHandles = 0;
			$this->openedFileName = $filename;
		}
		
		return $ret;
	}
	
	/**
	 * Close the active archive (opened or newly created)
	 * @link http://php.net/manual/en/function.ziparchive-close.php
	 * @return bool Returns true on success or false on failure.
	 */
	public function close () {
		$this->openedFileName = null;
		return parent::close();
	}
	
	/**
	 * Reopens the currently open zip archive
	 * @return TRUE if closing and opening was successful. FALSE in all other
	 * case, including if there is nothng to reopen.
	 */
	public function reopen() {
		if ($this->openedFileName === null)
			return false;
		
		$openedFilenameBackup = $this->openedFileName; // will be erased by close()
		
		if ($this->close() === false)
			return false;
		
		return $this->open($openedFilenameBackup);
	}
	
	/**
	 * Adds a file to a ZIP archive from the given path
	 * @param filename string The path to the file to add.
	 * @param localname string[optional] local name inside ZIP archive.
	 * @return bool Returns true on success or false on failure.
	 */
	public function addFile($filename, $localname = null) {
		if ($this->activeHandles > self::REOPEN_EVERY) {
			$this->reopen();
		}
		
		$this->activeHandles++;
		return parent::addFile($filename, $localname);
	}
	
	/** Recursively adds a directory to a zip archive.
	 *
	 * @param string $dir The path to the directory to add
	 * @param string $base The path where to put the directory in the zip
	 * (defaults to the name of the directory to add)
	 */
	public function addDirectory($dir, $base = null) {
		if (! is_dir($dir))
			throw new ZipException('Cannot add non-dir to zip');
		
		if ($base == null)
			$base = basename($dir);
		$this->addEmptyDir($base);

        foreach (new DirectoryIterator($dir) as $entry) {
        	if ($entry->isDot())
        		continue;
        	
            if ($entry->isFile())
            	$this->addFile($entry->getPathname(), $base.'/'.$entry->getFilename());
            else if ($entry->isDir())
            	$this->addDirectory($entry->getPathname(), $base.'/'.$entry->getFilename());
            else
            	trigger_error('Skipped an entry that is neither a file nor a directory during zipping', E_NOTICE);
        }
	}
}

class ZipException extends Exception {}
?>
