<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
class PhpThumbComponent extends Object {
	public $components = array('File');
	
	private static $CACHE;
	private static $CACHE_SUBDIR = 'thumbs';
	private static $MAX_AGE = 1296000; // 15 days

	public function startup(&$controller) {
		// I hate this way of not letting expressions as default values
		self::$CACHE = IMAGES.self::$CACHE_SUBDIR.DS;
		
		// Check cache directory
		if (! file_exists(self::$CACHE)) {
			if (! mkdir(self::$CACHE)) {
				trigger_error('Cannot create '.self::$CACHE, E_USER_ERROR);
				return;
			}
		} else if (! is_dir(self::$CACHE)) {
			trigger_error(self::$CACHE.' is not a directory', E_USER_ERROR);
			return;
		}
		if (! is_writable(self::$CACHE)) {
			trigger_error(self::$CACHE.' is not writable', E_USER_ERROR);
			return;
		}
		
		// Prune
		$dh = opendir(self::$CACHE);
		while (($file = readdir($dh)) !== false) {
            if (is_file(self::$CACHE.$file)
				&& filemtime(self::$CACHE.$file)+self::$MAX_AGE < time())
				if (! unlink(self::$CACHE.$file))
					$this->log('Failed to delete obsolete thumb file '.$file);
        }
		closedir($dh);
		
		App::import('Vendor', 'phpThumb', array('file' => 'phpThumb'.DS.'phpthumb.class.php'));
	}
	
	/** Get the URL to a thumbnail located in the img directory. The thumbnail
	 * is not guaranteed to be already generated, but if it isn't, a call to
	 * getThumbnailFromURL will generate it.
	 * @param $image The image URL, either absolute or relative to the IMAGES
	 * directory.
	 * @return The thumbnail path, relative to the IMAGES_URL directory
	 */
	public function getThumbnail($image, $width = 0, $height = 0) {
		list($image, $isRemote) = $this->checkImage($image);
		
		if ($width <= 0)
			$width = Configure::read('Demyo.maxThumbWidth');
		if ($height <= 0)
			$height = Configure::read('Demyo.maxThumbHeight');
		
		if ($isRemote)
			return $this->getAnonymousThumbnail($image, $width, $height);
		else
			return $this->getDelayedThumbnail($image, $width, $height);
	}
	
	/** Checks if the image is local, and change its patch relative to the
	 * images directory if it is.
	 * @param $image The image URL
	 * @return array An array. The first element is the fixed image path (can
	 * be the unchanged parameter). The second element is a boolean indicating
	 * wether the image is remote.
	 */
	private function checkImage($image) {
		$isRemote = (preg_match('|^https?://|', $image, $matches) === 1);
		
		if (! $isRemote) {
			/*$image = IMAGES.$image;
			if (! is_file($image)) {
				trigger_error($image.' is not a file', E_USER_WARNING);
				return false;
			}*/
		}
		
		return array($image, $isRemote);
	}
	
	/** Renders a thumbnail in a file, and returns the url to the thumbnail.
	 * @param width The maximum thumbnail width. Any value below 0 will be
	 * replaced by the configured maximum size.
	 * @param height The maximum thumbnail height. Any value below 0 will be
	 * replaced by the configured maximum size.
	 * @return The thumbnail path, relative to the IMAGES_URL directory
	 */
	private function getAnonymousThumbnail($image, $width, $height) {
		$outputFilename = md5($image.$width.$height).'.jpg';
		if (file_exists(self::$CACHE.$outputFilename))
			return self::$CACHE_SUBDIR.'/'.$outputFilename;
		
		return $this->generateThumbnail($image, $outputFilename, 'jpg', $width, $height);
	}
	
	
	/**
	 * Returns the URL to reach the thumbnail of an image.
	 * This method does not actually generate the thumbnail. It is intended
	 * as a way to make the URL that will be passed to getThumbnailFromURL()
	 * at a later time.
	 *
	 * @param width The maximum thumbnail width. Any value below 0 will be
	 * replaced by the configured maximum size.
	 * @param height The maximum thumbnail height. Any value below 0 will be
	 * replaced by the configured maximum size.
	 * @return The thumbnail path, relative to the IMAGES_URL directory (so that
	 * $html->image() can be used).
	 */
	private function getDelayedThumbnail($image, $width, $height) {
		$info = pathinfo($image);
		if (! empty($info['dirname']))
			$info['dirname'] .= '/';
		return self::$CACHE_SUBDIR
			."/{$info['dirname']}{$info['filename']}_{$width}x{$height}.{$info['extension']}";
	}
	
	private function generateThumbnail($source, $destination, $format, $width, $height) {
		$phpThumb = new phpThumb();
		
        $phpThumb->setSourceFilename($source);
		$phpThumb->setParameter('config_allow_src_above_docroot', true);
        $phpThumb->setParameter('w',$width);
        $phpThumb->setParameter('h',$height);
        
		if ($format == 'png') {
			$phpThumb->setParameter('config_output_format', 'png');
		} else {
			$phpThumb->setParameter('q',75);
			$phpThumb->setParameter('config_output_format', 'jpeg');
		}
		
		if($phpThumb->generateThumbnail()){
			if(!$phpThumb->RenderToFile(self::$CACHE.$destination)) {
				$this->cakeError('error500', array('code' => '500', 'name' => __('Error', true), 'message' => 'Failed to generate thumbnail for '.basename($source). ' : '.$phpThumb->fatalerror.(implode("\n*", $phpThumb->debugmessages))));
				return false;
			}
			
			if (strtoupper(substr(PHP_OS, 0, 3)) !== 'WIN')
				chmod(self::$CACHE.$destination, 0666); // See bug #91
		} else {
			$this->cakeError('error500', array('code' => '500', 'name' => __('Error', true), 'message' => 'Failed to generate thumbnail for '.basename($source). ' : '.$phpThumb->fatalerror.(implode("\n*", $phpThumb->debugmessages))));
			return false;
		}

		return self::$CACHE_SUBDIR.'/'.$destination;
	}
	
	/**
	 * Generates a thumbnail corresponding to a given URL.
	 * If the thumbnail does not exist, it will be generated and cached in
	 * the thumbs directory, so that further direct calls to the file will be
	 * possible.
	 * @param url string A string of the form foo/bar/image_[width]x[height].ext
	 * @return The URL to the generated thumbnail, which is usually (but not
	 * always) the passed parameter, prefixed with the thumbnails cache
	 * directory. Direct calls to redirect() can be made for this URL.
	 */
	public function getThumbnailFromURL($url) {
		if (preg_match('|^(.*/)?([^/]+)_(\d+)x(\d+)\.([^.]+)$|', $url, $matches) === 0)
			throw new Exception("$url is not a valid file to check for");
		if (DS !== '/')
			$matches[1] = str_replace('/', DS, $matches[1]);
		
		$width = (int)$matches[3];
		$height = (int)$matches[4];
		$outputDirRelative = $matches[1];
		$outputDirAbsolute = self::$CACHE.$matches[1];
		$extension = $matches[5];
		$normalizedExt = strtolower($extension);
		$sourceFileRelative = "{$outputDirRelative}{$matches[2]}.{$extension}";
		$sourceFileAbsolute = IMAGES.DS.$sourceFileRelative;
		
		// Build output file, see if it doesn't exist already
		if (! in_array($normalizedExt, array('png', 'jpg', 'jpeg'))) {
			$extension = 'jpg';
			$normalizedExt = 'jpg';
		}
		$outputFileRelative = "{$outputDirRelative}{$matches[2]}_{$width}x{$height}.$extension";
		$outputFileWWW = '/'.IMAGES_URL.self::$CACHE_SUBDIR.'/'.$outputFileRelative;
		$outputFileAbsolute = self::$CACHE.$outputFileRelative;
		if (file_exists($outputFileAbsolute))
			return $outputFileWWW;

		if (! file_exists($sourceFileAbsolute))
			$this->cakeError('error404');
		if ($this->File->dir_cmp(IMAGES, $sourceFileAbsolute) < 0)
			$this->cakeError('error403');
		
		// Create subdirectories if they don't exist
		$mask = umask(0); // Allow write access to everyone
		if (! file_exists($outputDirAbsolute))
			mkdir($outputDirAbsolute, 0777, true);
		if (! is_dir($outputDirAbsolute))
			$this->cakeError('error500', array('message' => 'Output directory is actually a file'));
		
		$this->generateThumbnail($sourceFileAbsolute, $outputFileRelative, $normalizedExt, $width, $height);

		umask($mask);
		return $outputFileWWW;
	}
	
	/** Clears the thumbnails directory */
	public function clean() {
		$this->File->rm(self::$CACHE, true);
	}
}

?>