<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
class Album extends AppModel {
	public $name = 'Album';
	public $actsAs = array('Containable', 'SimpleXML' => array('useHABTM' => true));
	public $hasMany = array('Derivative',
		// Actually dependent, but handled by the database.
		// Cake does not handle columns with compounf key. See Demyo bug #264
		'AlbumPrice' => array('order' => 'date ASC', 'dependent' => false, 'exclusive' => true)
	);
	public $belongsTo = array('Publisher', 'Collection', 'Binding', 'Series',
		/* An album doesn't actually belong to a cover, but we want the ID to
		 * be part of Album and not of Cover (which doesn't exist and is an alias
		 * to Image)
		 */
		'Cover' => array('className' => 'Image'/*, 'dependent' => false*/)); // Dependency is handled manually since Cake want an 'id' field
	public $hasAndBelongsToMany = array(
		'Artist' =>
			array('className' => 'Artist',
				'joinTable' => 'albums_artists',
				'foreignkey' => 'album_id',
				'associationForeignKey' => 'artist_id',
				'order'=>'name ASC',
				'uniq' => true,
			),
		'Writer' =>
			array('className' => 'Writer',
				'joinTable' => 'albums_writers',
				'foreignkey' => 'album_id',
				'associationForeignKey' => 'writer_id',
				'order'=>'name ASC',
				'uniq' => true,
			),
		'Colorist' =>
			array('className' => 'Colorist',
				'joinTable' => 'albums_colorists',
				'foreignkey' => 'album_id',
				'associationForeignKey' => 'colorist_id',
				'order'=>'name ASC',
				'uniq' => true,
			),
		'Tag' =>
			array('className' => 'Tag',
				'order' => 'Tag.name ASC',
				'uniq' => true,
			),
		'Image' => array(
			'className' => 'Image',
			'order' => 'description ASC'),
		'Borrower' => array(
			'className' => 'Borrower',
			'conditions' => array('OR' => array('AlbumsBorrower.return_date IS NULL', 'AlbumsBorrower.return_date' => '0000-00-00')),
			'with' => 'AlbumsBorrower'
		)
	);
	public $validate = array(
		'series_id' => array('rule' => 'notEmpty', 'required' => true),
		'publisher_id' => array('rule' => 'notEmpty', 'required' => true),
		'title' => array('rule' => 'notEmpty'),
		'cycle' => array('rule' => array('extended_number', false, 0, 0xFFFF), 'allowEmpty' => true),
		'number' => array('rule' => array('extended_number', true, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		// number_suffix is limited by the size of the input field
		'first_edition' => array('rule' => array('date', 'ymd'), 'allowEmpty' => true),
		'this_edition' => array('rule' => array('date', 'ymd'), 'allowEmpty' => true),
		'isbn' => array ('rule' => array('custom', '/^[0-9-]*X?$/i')),
		'purchase_date' => array('rule' => array('date', 'ymd'), 'allowEmpty' => true),
		'purchase_price' => array('rule' => array('extended_number', true, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		'height' => array('rule' => array('extended_number', true, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		'width' => array('rule' => array('extended_number', true, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		'pages' => array('rule' => array('extended_number', false, 0, 0x7FFFFFFF), 'allowEmpty' => true)
	);
	
	public function beforeValidate() {
		$this->validate['series_id']['message'] = __('The series cannot be empty', true);
		$this->validate['publisher_id']['message'] = __('The publisher cannot be empty', true);
		$this->validate['title']['message'] = __('The title cannot be empty', true);
		$this->validate['cycle']['message'] = __('This is not a valid positive number', true);
		$this->validate['number']['message'] = __('This is not a valid positive decimal number', true);
		$this->validate['first_edition']['message'] = __('This date is invalid', true);
		$this->validate['this_edition']['message'] = __('This date is invalid', true);
		$this->validate['isbn']['message'] = __('This ISBN is invalid', true);
		$this->validate['purchase_date']['message'] = __('This date is invalid', true);
		$this->validate['purchase_price']['message'] = __('This is not a valid positive decimal number', true);
		$this->validate['height']['message'] = __('This is not a valid positive decimal number', true);
		$this->validate['width']['message'] = __('This is not a valid positive decimal number', true);
		$this->validate['pages']['message'] = __('This is not a valid positive number', true);
	}
	
	/**
	 * Removes record for given ID. If no ID is given, the current ID is used. Returns true on success.
	 *
	 * @param mixed $id ID of record to delete
	 * @param boolean $cascade Set to true to delete records that depend on this record
	 * @return boolean True on success
	 * @access public
	 * @link http://book.cakephp.org/view/690/del
	 */
	function del($id = null, $cascade = true) {
		if (!empty($id)) {
			$this->id = $id;
		}
		$id = $this->id;
		
		// Remove price history. Fake it as not existing when running the parent.
		$this->AlbumPrice->deleteForAlbum($id);
		$backup = $this->hasMany['AlbumPrice'];
		unset($this->hasMany['AlbumPrice']);
		$ret = parent::del($id, $cascade);
		$this->hasMany['AlbumPrice'] = $backup;
		return $ret;
	}
	
	public function getImages($id) {
		$this->unbindAll(array('belongsTo' => array('Cover'), 'hasAndBelongsToMany' => array('Image')));
		$album = $this->read(null, $id);
		return array_merge(array($album['Cover']), $album['Image']);
	}
	
	public function getAlbumCount($model, $id) {
		if (! isset($this->belongsTo[$model])) {
			$this->log("$model is not a associated to {$this->name}", LOG_WARNING);
			return 0;
		}
		
		$model = Inflector::underscore($model);
		return $this->find('count', array(
			'conditions' => array("{$this->name}.{$model}_id" => $id),
			'recursive' => -1
		));
	}
	
	public static function sortWithSeries($a, $b) {
		if ($a['Series']['name'] != $b['Series']['name'])
			return ($a['Series']['name'] > $b['Series']['name']) ? +1 : -1;

		if ($a['cycle'] != $b['cycle'])
			return ($a['cycle'] > $b['cycle']) ? +1 : -1;
		
		if ($a['number'] != $b['number'])
			return ($a['number'] > $b['number']) ? +1 : -1;
		
		if ($a['number_suffix'] != $b['number_suffix'])
			return ($a['number_suffix'] > $b['number_suffix']) ? +1 : -1;
		
		return 0;
	}
	
	public function afterSave($created) {
		// Since the class having the association is actually a subClass of Author
		// the automatic cache clearance routing won't clear the associations
		// (this::belongsTo et al. don't link to author).
		// Let's just do it by ourselves
		clearCache(array('authors'));
	}
	
	public function getEditionsPerYear() {
		// Reminder: we sort DESC because filtering the top N values is then
		// much easier (else we'd have to extract the values from the end
		// of the array and unshift them in a result array
		// It makes sense anyway: the most recent data first (usually, that's
		// what people want)
		return Set::extract('/0/.',
			$this->query("SELECT YEAR(this_edition) AS year, COUNT(*) AS count "
				."FROM {$this->tablePrefix}{$this->table} GROUP BY year ORDER BY year DESC"));
	}
	
	public function getPurchaseStatisticsByYear() {
		return Set::extract('/0/.',
			$this->query("SELECT YEAR(purchase_date) AS year ,COUNT(*) AS count "
				."FROM {$this->tablePrefix}{$this->table} GROUP BY year ORDER BY year DESC"));
	}

	public function getPurchaseAverage($numYears = 200) {
		$result = $this->query('SELECT AVG(purchases) FROM (SELECT YEAR(purchase_date) AS year,'
			." COUNT(*) AS purchases FROM {$this->tablePrefix}{$this->table}"
			.' WHERE purchase_date != \'0000-00-00\' AND purchase_date IS NOT NULL'
			." GROUP BY year ORDER BY year DESC LIMIT $numYears) AS yearly_purchases");
		return $result[0][0]['AVG(purchases)'];
	}
	
	public function getPurchaseStatisticsForYear($year = false) {
		$year = (int)$year;
		if ($year <= 0) { // Auto-detect last year with data
			$year = $this->query("SELECT DISTINCT YEAR(purchase_date) AS year "
				."FROM {$this->tablePrefix}{$this->table} ORDER BY year DESC LIMIT 1");
			$year = $year[0][0]['year'];
		}
		
		return Set::extract('/0/.',
			$this->query("SELECT MONTH(purchase_date) AS month ,COUNT(*) AS count "
				."FROM {$this->tablePrefix}{$this->table} WHERE "
				."YEAR(purchase_date)='$year' GROUP BY month ORDER BY month DESC"));
	}
}

?>
