<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
class AppModel extends Model {
	
	/** Validation rule for numbers.
	 * 
	 * @param $allowFloats true if floats should be allowed (in non-scientific notation)
	 * @param $min The minimum allowed number
	 * @param $max The maximum allowed number
	 */
	public function extended_number($data, $allowFloats, $min, $max) {
		$number = current($data);
		
		if ($allowFloats && !preg_match('/^[-+]?[0-9]+(\\.[0-9]+)?$/', $number))
			return false;
		if (!$allowFloats && !preg_match('/^[-+]?[0-9]+$/', $number))
			return false;
		
		$number = intval($number);
		return ($min <= $number) && ($number <= $max);
	}
	
	/** Validation rule for hexadecimal colors (with leading #).
	 */
	public function hex_color($data, $allowShort) {
		$color = current($data);
		
		if ($allowShort)
			$allowShort = 1;
		else
			$allowShort = 2;
		
		return preg_match("/^#([A-Z0-9]{3}){{$allowShort},2}$/i", $color);
	}
	
	public function find($conditions = null, $fields = array(), $order = null, $recursive = null) {
		/* find('orphans', array('associations' => array('Model1', 'Model2'))
		 * finds records that aren't associated with any record of the given models.
		 * If the model isn't associated with one of the models, the result
		 * set is considered to be empty.
		 */
		if ($conditions === 'orphans') {
			if (! isset($fields['associations']))
				return array();
			
			$ids = null;
			
			$name = $this->name;
			foreach (array('hasOne', 'hasMany', 'belongsTo', 'hasAndBelongsToMany') as $type) {
				foreach($this->$type as $model => $relation) {
					if (! in_array($model, $fields['associations']))
						continue;
					
					switch($type) {
						case 'belongsTo':
							$belongIDs = Set::extract("/$model/id",
								$this->$model->find('all', array('fields' => 'id', 'recursive' => -1)));
							$tempResult = Set::extract("/$name/id", $this->find('all',
								array('fields' => 'id', 'recursive' => -1, 'conditions' =>
									"$name.{$relation['foreignKey']} NOT IN (".implode(',', $belongIDs).')')));
							break;
						
						case 'hasMany':
							$prefix = $this->tablePrefix;
							$tempResult = Set::extract("/$name/id", $this->query("SELECT $name.id FROM $prefix{$this->table} $name"
								." LEFT JOIN $prefix{$this->$model->table} Other ON $name.id=Other.{$relation['foreignKey']}"
								." GROUP BY $name.id HAVING COUNT(Other.id) = 0"));
							break;
						
						case 'hasAndBelongsToMany':
							$prefix = $this->tablePrefix;
							$tempResult = Set::extract("/$name/id", $this->query("SELECT $name.id FROM $prefix{$this->table} $name"
								." LEFT JOIN $prefix{$relation['joinTable']} Other ON $name.id=Other.{$relation['foreignKey']}"
								." GROUP BY $name.id HAVING COUNT({$relation['associationForeignKey']}) = 0"));
							break;
						
						default:
							trigger_error("Unhandled association type for orphans: $type", E_USER_ERROR);
					}
					
					if ($ids === null)
						$ids = $tempResult;
					else
						$ids = array_intersect($ids, $tempResult);
				}
			}
			
			if (! empty($ids)) {
				$this->unbindAll();
				return $this->find('all', array('conditions' => array("$name.id" => $ids)));
			}
			else
				return array();
		}
		
		return parent::find($conditions, $fields, $order, $recursive);
	}
	
	public function findPageNumberStartsWith($field, $from, $limit, $conditions = array()) {
		$countBefore = $this->find('count', array('conditions' => array_merge($conditions, array("$field <" => $from))));
		return floor(1 + $countBefore / $limit);
	}
	
	/** Returns the name of the selected record, as defined by
	 * Model::displayField
	 *
	 * @param integer $id The ID of the record
	 * @return string
	 */
	public function getName($id = null) {
		if ($id === null)
			$id = $this->id;
		
		$data = $this->find('first', array(
			'fields' => array($this->displayField),
			'conditions' => array($this->name.'.id' => $id),
			'recursive' => -1
		));
		
		return $data[$this->name][$this->displayField];
	}
	
	/** Unbinds all associations buts the specified ones.
	 * Parameters are similar to those of unbindModel()
	 * @param array $exceptions And array of associations to keep
	 * @param boolean $reset True if the associations should be reset
	 */
	public function unbindAll($exceptions = array(), $reset = true) {
		$actualUnbind = array();
		
		foreach (array('hasOne', 'hasMany', 'belongsTo', 'hasAndBelongsToMany') as $type) {
			if (! isset($this->$type))
				continue;
			
			$actualUnbind[$type] = array();
			
			if (! isset($exceptions[$type])) { // Don't check everything one by one
				$actualUnbind[$type] = array_keys($this->$type);
				continue;
			}
			
			foreach(array_keys($this->$type) as $assoc) {
				if (! in_array($assoc, $exceptions[$type]))
					array_push($actualUnbind[$type], $assoc);
			}
		}
		
		return $this->unbindModel($actualUnbind, $reset);
	}

	public function onError() {
		$db =& ConnectionManager::getDataSource($this->useDbConfig);
		$err = $db->lastError();
		if (preg_match('/^1451:/', $err)) { // Cannot delete or update a parent row: a foreign key constraint fails
			// Nice, but a pain to localize
			/*preg_match("/{$this->tablePrefix}([^`]+)`, CONSTRAINT .* REFERENCES `{$this->tablePrefix}([^`]+)`/", $err, $matches);
			$constrainer = __(strtolower(Inflector::humanize(Inflector::singularize($matches[1]))), true);
			$deleter = __(strtolower(Inflector::humanize(Inflector::singularize($matches[2]))), true);
			$this->errorDetails = __s('This %1$s cannot be deleted because at least one %2$s depends on it. Remove the dependencies and try again.', $deleter, $constrainer);*/
			$this->errorDetails = __('This item cannot be deleted because at least one other depends on it. Remove the dependencies and try again.', true);
		}
	}
}
?>