<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
class Derivative extends AppModel {
	public $name = 'Derivative';
	public $actsAs = array('Containable', 'SimpleXML' => array('useHABTM' => true));
	public $belongsTo = array('Series', 'Album', 'DerivativeType', 'Artist', 'Source');
	public $hasMany = array(
		// Actually dependent, but handled by the database.
		// Cake does not handle columns with compounf key. See Demyo bug #264
		'DerivativePrice' => array('order' => 'date ASC', 'dependent' => false, 'exclusive' => true)
	);
	public $hasAndBelongsToMany = array(
		'Image' => array(
			'className' => 'Image',
			'order' => 'description ASC')
	);
	public $validate = array(
		'series_id' => array('rule' => 'notEmpty', 'required' => true),
		'derivative_type_id' => array('rule' => 'notEmpty', 'required' => true),
		'source_id' => array('rule' => array('comparison', '>=', 1)),
		'colors' => array('rule' => array('extended_number', false, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		'number' => array('rule' => array('extended_number', false, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		'total' => array('rule' => array('extended_number', false, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		'width' => array('rule' => array('extended_number', true, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		'height' => array('rule' => array('extended_number', true, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		'depth' => array('rule' => array('extended_number', true, 0, 0x7FFFFFFF), 'allowEmpty' => true),
		'purchase_date' => array('rule' => array('date', 'ymd'), 'allowEmpty' => true),
		'purchase_price' => array('rule' => array('extended_number', true, 0, 0x7FFFFFFF), 'allowEmpty' => true),
	);
	
	public function beforeValidate() {
		$this->validate['series_id']['message'] = __('The series cannot be empty', true);
		$this->validate['derivative_type_id']['message'] = __('The derivative type cannot be empty', true);
		$this->validate['source_id']['message'] = __('The source cannot be empty', true);
		$this->validate['colors']['message'] = __('This is not a valid positive number', true);
		$this->validate['number']['message'] = __('This is not a valid positive number', true);
		$this->validate['total']['message'] = __('This is not a valid positive number', true);
		$this->validate['width']['message'] = __('This is not a valid positive decimal number', true);
		$this->validate['height']['message'] = __('This is not a valid positive decimal number', true);
		$this->validate['depth']['message'] = __('This is not a valid positive decimal number', true);
		$this->validate['purchase_date']['message'] = __('This date is invalid', true);
		$this->validate['purchase_price']['message'] = __('This is not a valid positive decimal number', true);
	}
	
	/**
	 * Removes record for given ID. If no ID is given, the current ID is used. Returns true on success.
	 *
	 * @param mixed $id ID of record to delete
	 * @param boolean $cascade Set to true to delete records that depend on this record
	 * @return boolean True on success
	 * @access public
	 * @link http://book.cakephp.org/view/690/del
	 */
	function del($id = null, $cascade = true) {
		if (!empty($id)) {
			$this->id = $id;
		}
		$id = $this->id;
		
		// Remove price history. Fake it as not existing when running the parent.
		$this->DerivativePrice->deleteForDerivative($id);
		$backup = $this->hasMany['DerivativePrice'];
		unset($this->hasMany['DerivativePrice']);
		$ret = parent::del($id, $cascade);
		$this->hasMany['DerivativePrice'] = $backup;
		return $ret;
	}
	
	/* After find callback, builds a meta_name based on the type, source and
	 * Series.
	 *
	 * @param mixed $results The results of the find operation
	 * @param boolean $primary Whether this model is being queried directly (vs. being queried as an association)
	 * @return mixed Result of the find operation
	 * @access public
	 */
	public function afterFind($results, $primary = false) {
		foreach ($results as &$result) {
			if (isset($result[$this->name])) { // result was ['Model' => ...]
				$data =& $result[$this->name];
			}
			else if (isset($result['source_id'])) // result was ['field1', 'field2']
				$data =& $result;
			else
				continue;
			
			$series = '';
			if (isset($data['Series']) && isset($data['Series']['name']))
				$series = $data['Series']['name'];
			else if (isset($result['Series']) && isset($result['Series']['name']))
				$series = $result['Series']['name'];
			
			$type = '';
			if (isset($data['DerivativeType']) && isset($data['DerivativeType']['name']))
				$type = $data['DerivativeType']['name'];
			else if (isset($result['DerivativeType']) && isset($result['DerivativeType']['name']))
				$type = $result['DerivativeType']['name'];
			
			$source = '';
			if (isset($data['Source']) && isset($data['Source']['name']))
				$source = $data['Source']['name'];
			else if (isset($result['Source']) && isset($result['Source']['name']))
				$source = $result['Source']['name'];

			if (empty($data)) // no matching record
				continue;
			
			if (! is_array(current($data))) { // result was ['Model' => ['field1', 'field2']]
				if ($series != '' && $type != '') { // $type is mandatory, but could be not part of the selected fields for this query
					if ($source != '')
						$data['meta_name'] = __s('%1$s for %2$s from %3$s', $type, $series, $source);
					else
						$data['meta_name'] = __s('%1$s for %2$s', $type, $series);
				} else {
					// Be nice with orphaned entries. Define the meta_name index
					// so that it can be used elsewhere without checks.
					$data['meta_name'] = '';	
				}
			}
			// else result was ['Model' => [0 => ['field1', 'field2']]], already
			// handled in a previous method call (the call for the sub-items)
		}
		
		return $results;
	}
	
	public function getDerivativesCount($model, $id) {
		if (! isset($this->belongsTo[$model])) {
			$this->log("$model is not a associated to {$this->name}", LOG_WARNING);
			return 0;
		}
		
		$model = Inflector::underscore($model);
		return $this->find('count', array(
			'conditions' => array("Derivative.{$model}_id" => $id),
			'recursive' => -1
		));
	}
	
	public function afterSave($created) {
		// Since the class having the association is actually a subClass of Author
		// the automatic cache clearance routing won't clear the associations
		// (this::belongsTo et al. don't link to author).
		// Let's just do it by ourselves
		clearCache(array('authors'));
	}
}

?>
