<?php
/* Copyright 2007-2012 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
class Search extends AppModel {
	const STATUS_TEMP = 'temporary';
	const STATUS_PERMANENT = 'permanent';
	const DATE_FORMAT = 'Y-m-d H:i:s';
	
	public $name = 'Search';
	public $actsAs = array('SimpleXML' => array('useHABTM' => true));
	
	/**
	 * Saves a temporary search.
	 * @param string $model The type of concept being searched.
	 * @param string $criteria The JSON criteria
	 * @param array $parameters An array of parameters to save in the query.
	 */
	public function saveTemporary($model, $criteria, $parameters) {
		$data = array('Search' => array(
			'model' => $model,
			'status' => Search::STATUS_TEMP,
			'criteria' => $criteria,
			'parameters' => json_encode($parameters),
			'last_used' => date(Search::DATE_FORMAT)
		));
		$this->save($data);
	}
	
	public function refreshUsage($id) {
		$this->id = $id;
		$this->saveField('last_used', date(Search::DATE_FORMAT));
	}
	
	/**
	 * Converts a JSON criteria to conditions usable for a search in CakePHP.
	 * @param string $criteria The criteria, in JSON, to convert.
	 */
	public function decodeCriteria($criteria) {
		return $this->jsonObjectToCakeCriteria(json_decode($criteria));
	}
	
	public function decodeParameters($parameters) {
		if (empty($parameters)) {
			return array();
		}
		return (array) json_decode($parameters);
	}

	private function jsonObjectToCakeCriteria($criteria) {
		$conditions = array();

		if (!empty($criteria->or)) {
			$conditions['or'] = $this->jsonObjectToCakeCriteria($criteria->or);
		} else if (!empty($criteria->and)) {
			$conditions['and'] = $this->jsonObjectToCakeCriteria($criteria->and);
		} else if (is_array($criteria)) {
			foreach ($criteria as $criteriaLine) {
				$conditions[] = $this->jsonObjectToCakeCriteria($criteriaLine);
			}
		} else if (is_object($criteria)) {
			$value = $criteria->value;
			$operator = $criteria->operator;
			if ($operator == 'LIKE' || $operator == 'NOT LIKE') {
				$value = "%$value%";
			} else if ($operator == 'IS NULL' || $operator == 'IS NOT NULL') {
				$value = null;
			}
			if ($value != null) {
				$conditions["{$criteria->field} $operator"] = $value;
			} else {
				array_push($conditions, "{$criteria->field} $operator");
			}
		}
		
		return $conditions;
	}

	/**
	 * When saving a search, always prune the old ones.
	 * @see Model::afterSave()
	 */
	public function afterSave($created) {
		$timeout = Configure::read('Demyo.Search_temp_timeout') * 3600; // Convert to seconds
		$this->deleteAll(array('status' => Search::STATUS_TEMP, 'last_used < ' => date(Search::DATE_FORMAT, time() - $timeout)));
	}
}

?>
