<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
class Series extends AppModel {
	public $name = 'Series';
	public $order = 'name ASC';
	public $actsAs = array('Containable', 'SimpleXML' => array('useHABTM' => true));
	public $hasMany = array(
		'Album'	=> array(
			'className' => 'Album',
			'order' => 'Album.cycle ASC, Album.number ASC, Album.number_suffix ASC',
			'dependent' => false
			),
		'Derivative');
	public $hasAndBelongsToMany = array(
		'RelatedSeries' => array(
			'className' => 'Series',
			'joinTable' => 'series_relations',
			'foreignKey' => 'main',
			'associationForeignKey' => 'sub',
			'order' => 'RelatedSeries.name'
			)
		);
	public $validate = array(
		'name' => array('rule' => 'notEmpty'),
		'website' => array('rule' => 'url', 'allowEmpty' => true)
	);
	
	public function beforeValidate() {
		$this->validate['name']['message'] = __('The name cannot be empty', true);
		$this->validate['website']['message'] = __('This is not a valid address', true);
	}

	/**
	 * Saves model data to the database. By default, validation occurs before save.
	 *
	 * @param array $data Data to save.
	 * @param boolean $validate If set, validation will be done before the save
	 * @param array $fieldList List of fields to allow to be written
	 * @return mixed On success Model::$data if its not empty or true, false on failure
	 */
	public function save($data = null, $validate = true, $fieldList = array()) {
		$main = $this->hasAndBelongsToMany['RelatedSeries']['foreignKey'];
		$sub = $this->hasAndBelongsToMany['RelatedSeries']['associationForeignKey'];
		$table = $this->tablePrefix.$this->hasAndBelongsToMany['RelatedSeries']['joinTable'];
		
		/*
		We have to edit the reversed relation, i.e. set the other series
		to have a relation with the current. The relation of the current
		series with the other ones is handled by parent::save.

		The easiest way is to delete all reversed relations and re-create them.
		It's not the most efficient but it is reliable. Computing differences
		was implemented until Demyo 1.4 but caused problems at import for
		obscure reasons (the current set of reversed relations couldn't be
		computed reliably, which led to duplicate relations).
		*/
		
		if (! empty($data['Series']['id'])) { // edition of an existing series
			if (! ctype_digit($data['Series']['id']))
				return false;
			$id = $data['Series']['id'];
			$result = parent::save($data, $validate, $fieldList);
			if (! $result)
				return $result;
		}
		
		else { // insertion of a new series
			$result = parent::save($data, $validate, $fieldList);
			if (! $result)
				return $result;
			$id = $this->id;
		}
		
		// The following does not depend on the add/edit action anymore
		$relations = (! empty($data['RelatedSeries']['RelatedSeries']))?$data['RelatedSeries']['RelatedSeries']:array();

		$this->query("DELETE FROM $table WHERE $sub = $id");
		if (! empty($relations)) {
			foreach ($relations as &$item)
				$item = "($item, $id)";
			$this->query("INSERT INTO $table ($main, $sub) VALUES ".implode(",", $relations));
		}
		
		return $result;
	}
	
	/**
	 * Removes record for given id. If no id is given, the current id is used.
	 *
	 * @param mixed $id ID of record to delete
	 * @param boolean $cascade Set to true to delete records that depend on this record
	 * @return boolean True on success
	 * @access public
	 */
	function del($id = null, $cascade = true) {
		if (! parent::del($id, $cascade))
			return false;
		
		// We also need to delete reverse relations since they were inserted
		// manually
		$sub = $this->hasAndBelongsToMany['RelatedSeries']['associationForeignKey'];
		$table = $this->tablePrefix.$this->hasAndBelongsToMany['RelatedSeries']['joinTable'];
		$this->query("DELETE FROM $table WHERE $sub = $id");
		return true;
	}
		
	public function getImages($id = null) {
		if ($id === null)
			$id = $this->id;
		
		$this->recursive = 2;
		$this->Album->unbindAll(array('belongsTo' => array('Cover'), 'hasAndBelongsToMany' => array('Image')));
		$this->Derivative->unbindAll(array('hasAndBelongsToMany' => array('Image')));
		$series = $this->read(null, $id);
		
		$images = array();
		
		// Albums
		foreach ($series['Album'] as $album) {
			if (!empty($album['Cover'])) {
				$images = array_merge($images, array($album['Cover']), $album['Image']);
			} else {
				$images = array_merge($images, $album['Image']);
			}
		}
		
		// Derivatives
		foreach ($series['Derivative'] as $deriv)
			$images = array_merge($images, $deriv['Image']);
		
		return $images;
	}
	
	/** Adds a set of tags to all albums of a series. */
	public function tag($tagNames, $id = null) {
		if ($id === null)
			$id = $this->id;
		
		$tagIDs = $this->Album->Tag->getFromArray($tagNames);
		
		$this->Album->contain(array('Tag.id'));
		$albums = $this->Album->find('all', array('fields' => array('Album.id'),
			'conditions' => array('series_id' => $id)));
		foreach ($albums as &$album) {
			// Reminder: array_merge(array(1), array(1)) -> array(1, 1)
			// This is not acceptable since SQL will try to insert the same tuple twice
			$album['Tag'] = array_unique(array_merge(Set::extract('/id', $album['Tag']), $tagIDs));
		}
		$this->Album->saveAll($albums);
	}
	
	/** Removes a set of tags from all albums of a series. */
	public function unTag($tagNames, $id = null) {
		if ($id === null)
			$id = $this->id;
		
		$tagIDs = $this->Album->Tag->getFromArray($tagNames);
		
		$this->Album->contain(array('Tag.id'));
		$albums = $this->Album->find('all', array('fields' => array('Album.id'),
			'conditions' => array('series_id' => $id)));
		foreach ($albums as &$album) {
			$album['Tag'] = array_diff(Set::extract('/id', $album['Tag']), $tagIDs);
			if (empty($album['Tag']))
				$album['Tag'] = array(0 => false); // Else, Cake won't clear the existing ones
		}
		$this->Album->saveAll($albums);
	}
	
	// There's gotta be a more efficient way than making up to 3 queries...
	public function getSurroundings($seriesID, $albumID, $fields = array('id', 'title'), $getPrevious = true, $getNext = true) {
		// Get albums IDs
		$ids = $this->Album->find('all', array(
			'fields' => 'id',
			'order' => $this->hasMany['Album']['order'],
			'conditions' => array('series_id' => $seriesID),
			'recursive' => -1
		));
		
		// Find the index of the current one
		$i = 1;
		foreach ($ids as $id) {
			if ($id['Album']['id'] == $albumID)
				break;
			$i++;
		}
		
		$surroundings = array('previous' => null, 'next' => null);
		
		if ($i > 1) {
			$surroundings['previous'] = $this->Album->find('first', array(
				'fields' => $fields,
				'order' => $this->hasMany['Album']['order'],
				'conditions' => array('series_id' => $seriesID),
				'page' => $i-1,
				'recursive' => -1
			));
		}
		
		$potentialNext = $this->Album->find('first', array(
			'fields' => $fields,
			'order' => $this->hasMany['Album']['order'],
			'conditions' => array('series_id' => $seriesID),
			'page' => $i+1,
			'recursive' => -1
		));
		if (!empty($potentialNext))
			$surroundings['next'] = $potentialNext;
		
		return $surroundings;
	}
}

?>
