<?php
/* Copyright 2007-2009 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
class ChartHelper extends AppHelper {
	public $helpers = array('Html', 'jQuery');
	private $renderer;
	
	/** Sets up the chart renderer depending on user preferences */
	public function beforeRender() {
		$method = '';
		if (($method = Configure::read('Demyo.Stats_chartMethod')) === null)
			$method = 'Flash';
		
		switch ($method) {
			case 'Google':
				$this->renderer = new GoogleChartRenderer();
				break;
			case 'Flash':
				$this->renderer = new OpenFlashChartRenderer($this->Html);
				break;
			default:
				trigger_error('Invalid charting method: '.$method, E_USER_ERROR);
		}
		
		$this->renderer->beforeRender();
	}
	
	public function afterRender() {
		$this->renderer->afterRender();
	}
	
	/** Extracts data from a result set and returns a label set, a data set and
	 * a count of disregarded elements.
	 * 
	 * @param $labelIndex 
	 * @param $dataIndex
	 * @param $resultSet An array of the kind array(0 => array($labelIndex => X, $dataIndex => Y), 1 => ...)
	 * @param $maxResults The maximum number of entries to take into account.
	 * When this many entries have been counted, the rest of the result set is
	 * disregarded (the numbers in $dataIndex for diregarded entries are
	 * summed and returned).
	 * @return array(labels, data, notShownData)
	 */
	public function extract(&$resultSet, $labelIndex, $dataIndex, $maxResults) {
		$notShownData = 0;
		$labels = array();
		$data = array();
		
		// Flatten and restrict
		$i = 1;
		foreach($resultSet as $result) {
			if ($i++ <= $maxResults) {
				$labels[] = $result[$labelIndex];
				$data[] = $result[$dataIndex];
			} else
				$notShownData += $result[$dataIndex];
		}
		
		return array($labels, stringArrayToFloatArray($data), $notShownData);
	}
	
	/** Generates a chart with auto-selected renderer and returns HTML for it.
	 *
	 * @param data An array of data sets. Each set is itself an array of
	 * <em>numberic</em> values.
	 * @param labels An array of labels (indexed by 'x' and 'y'), preferably of
	 * the same length as the longest set
	 * @param params Additional optional params:
	 * x_grid and y_grid: to set the vertical and horizontal grid spacing in percent).
	 * x_label_orientation: horizontal, vertical, diagonal
	 * mark: a value => string pair to mark on the chart (for example, 50 => 'average')
	 * @param width Chart width in pixels
	 * @param height Chart height in pixels
	 * @return The HTML to include the chart.
	 */
	public function chart($data, $labels = array(), $params = array(), $width = -1, $height = -1) {
		if ($width == -1)
			$width = Configure::read('Demyo.Stats_chartWidth');
		if ($height == -1)
			$height = Configure::read('Demyo.Stats_chartHeight');

		if (! isset($params['type']))
			$params['type'] = 'bar';
		if (! isset($params['x_grid']))
			$params['x_grid'] = 0;
		if (! isset($params['y_grid']))
			$params['y_grid'] = 25;
		if (! isset($params['y_min']))
			$params['y_min'] = 0;
		
		return $this->renderer->chart($data, $labels, $width, $height, $params);
	}
}

abstract class ChartRenderer {
	const BG_COLOR = 'ffffff';
	const AXIS_COLOR = '666666';
	const GRID_COLOR = 'dddddd';
	const DATA_COLOR = '73a5d6';
	const MARK_COLOR = '73D6B3';
	/** Recommended bar width, incl. padding */
	const RECOM_BAR_WIDTH = 45;
	
	public function afterRender() {}
	public function beforeRender() {}
	public abstract function chart($data, $labels, $width, $height, $params);
}

class GoogleChartRenderer extends ChartRenderer {
	private static $CHART_URL = 'http://chart.apis.google.com/chart?';
	
	/** Generates a chart with Google's Chart API and returns an url
	 * for it.
	 *
	 * Maximum is detected automatically and y-scale will be set accordingly.
	 *
	 * Ampersands in the returned url are properly encoded.
	 *
	 * @return HTML for the Google chart image
	 */
	public function chart($data, $labels, $width, $height, $params) {
		$url = self::$CHART_URL;
		
		switch ($params['type']) {
			case 'line':
				$params['type'] = 'lc';
				break;
			case 'bar':
			default:
				$params['type'] = 'bvg';
		}
		
		$barcolor = self::DATA_COLOR;
		$bgcolor = self::BG_COLOR;
		$url.= "chco={$barcolor}&amp;chf=bg,s,{$bgcolor}";
		
		// We don't need a really large graph if we don't have a lot to show
		$maxLength = count2d($data);
		if ($width > $maxLength*self::RECOM_BAR_WIDTH)
			$width = $maxLength*self::RECOM_BAR_WIDTH;
		$url.= "&amp;chs={$width}x{$height}";
		
		// Type
		$url .= '&amp;cht='.$params['type'];
		
		// Normalize data
		$max = ceil(max2d($data)*1.05);
		$min = ($params['y_min'] === 'auto')?floor(min2d($data)*0.95):$params['y_min'];
		$ndata = $this->normalize($data, 100/($max-$min), $min);
		
		// Data sets
		$url .= '&amp;chd=t:';
		$i = 0; $ilen = sizeof($ndata);
		foreach ($ndata as $skey => $set) {
			$j = 0; $jlen = sizeof($set);
			foreach($set as $key => $value) {
				$url .= number_format($value, 1, '.', ''); // locale independent
				if ($j++ < $jlen-1)
					$url .= ',';
			}
			if ($i++ < $ilen-1)
				$url .= '|';
		}
		
		// Labels and mark
		$url .= '&amp;chxt=y';
		$hasX = isset($labels['x']);
		$hasM = isset($params['mark']);
		if ($hasX) $url .= ',x';
		if ($hasM) $url .= ',r';
		if ($hasX || $hasM) $url .= '&amp;chxl=';
		if ($hasX) {
			$url .= '1:';
			foreach($labels['x'] as $item)
				$url .= "|$item";
		}
		if ($hasM) {
			$axis = $hasX?2:1;
			$markAt = number_format(current(current($this->normalize(array(array(key($params['mark']))), 100/($max-$min), $min))), 1, '.', ''); // Normalized, formatted
			$markLabel = current($params['mark']);
			$url .= "|{$axis}:|{$markLabel}&amp;chxp={$axis},{$markAt}&chxs={$axis},".self::MARK_COLOR.',11,-1,t,'.self::MARK_COLOR."&amp;chxtc={$axis},-9001";
		}
		$url .= "&amp;chxr=0,$min,$max";
		
		// Grid
		if (isset($params['x_grid']) && isset($params['y_grid']))
			$url .= "&amp;chg={$params['x_grid']},{$params['y_grid']},1,5";
		
		return "<div class=\"chart\"><img src=\"$url\" alt=\"Google chart\" /></div>";
	}
	
	private function normalize($data, $factor, $base = 0) {
		$ret = array();
		
		foreach ($data as $skey => $set)
			foreach ($set as $key => $value)
				$ret[$skey][$key] = ($value-$base)*$factor;
		
		return $ret;
	}
}

class OpenFlashChartRenderer extends ChartRenderer {
	private $numCharts = 0;
	private $Html = null;
	const RECOM_BAR_WIDTH = 55;
	
	public function __construct(&$htmlHelper) {
		$this->Html = $htmlHelper;
		$this->Html->script('swfobject', false);
		$this->Html->script('json2', false);

		set_include_path(get_include_path().PATH_SEPARATOR.APP.'vendors'.DS.'OpenFlashChart');
		if (!App::import('Vendor', 'OFC_Chart', array('file' => 'OpenFlashChart'.DS.'OFC_Chart.php')))
			trigger_error('Failed to load OFC library', E_USER_ERROR);
	}
	
	/** Generates a chart with Open Flash Chart.
	 *
	 * @return HTML for the chart inclusion
	 */
	public function chart($data, $labels, $width, $height, $params) {
		$this->numCharts++;
		
		if (! isset($params['x_label_orientation']))
			$params['x_label_orientation'] = 'vertical';
		
		$maxLength = count2d($data);
		
		$chart = '';
		
		if ($maxLength == 2) // Bug 271: too little elements mess with the chart
			$maxLength++;
		// We don't need a really large graph if we don't have a lot to show
		if ($width > $maxLength*self::RECOM_BAR_WIDTH)
			$width = $maxLength*self::RECOM_BAR_WIDTH;
			
			// Print initialization functions on the first time
		if ($this->numCharts == 1) {
			$chart .= $this->Html->scriptBlock(
				"var chartData = {};
function ofc_get_data(id) {
	eval('var tmpData = chartData.'+id);
	return JSON.stringify(tmpData);
}"
			);
		}
		
		// Placeholder
		$chart .= "<div class=\"loading\" id=\"ofc-delayed-{$this->numCharts}\">Loading chart...</div>\n";
		
		// Create chart canvas
		$ofc = new OFC_Chart();
		$ofc->set_title(new OFC_Elements_Title());
		$ofc->set_bg_colour('#'.self::BG_COLOR);
		
		// Y Axis
		$y = new OFC_Elements_Axis_Y();
		$y->set_colour('#'.self::AXIS_COLOR);
		if ($params['y_min'] === 'auto') {
			if (isset($params['mark'])) {
				array_push($data, array(key($params['mark'])));
			}
			$min = floor(min2d($data)*0.95);
		} else {
			$min = $params['y_min'];
		}
		$max = ceil(max2d($data)*1.05);
		if ($params['y_grid'] == 0) { // Hide grid, but show labels
			$y->set_grid_colour('#'.self::BG_COLOR);
			$params['y_grid'] = 20;
		} else
			$y->set_grid_colour('#'.self::GRID_COLOR);
		$ySteps = (int)(($max-$min)*$params['y_grid']/100);
		// By setting the max a bit above the last label, we make sure it won't be rendered outside the chart
		$y->set_range($min, $max+($ySteps/2), $ySteps);
		// Sometimes, the value might be rendered outside the chart. The following is an ugly way to fix it.
		$y->labels->text = '   #val#';
		$ofc->set_y_axis($y);
		
		// X Axis
		$x = new OFC_Elements_Axis_X();
		$x->set_colour('#'.self::AXIS_COLOR);
		if ($params['x_grid'] == 0)
			$x->set_grid_colour('#'.self::BG_COLOR);
		else if ($params['x_grid'] > 0) // -1 to show on every step
			$x->set_steps($maxLength*$params['y_grid']/100);
		
		if (isset($labels['x'])) {
			$xLabels = new OFC_Elements_Axis_X_Label_Set();
			$xLabels->set_labels($labels['x']);
			$xLabels->rotate = $params['x_label_orientation'];
			$x->set_labels($xLabels);
		}
		$ofc->set_x_axis($x);
		
		$this->addDataToChart($ofc, $data, $params);
		
		if (isset($params['mark'])) {
			$mark = new OFC_Charts_Line();
			$mark->set_colour('#'.self::MARK_COLOR);
			$mark->{'dot-style'} = array ('type' => 'dot', 'dot-size' => 0, 'halo-size' => 0);
			$mark->set_values(array_fill(0, $maxLength, array('type' => 'dot', 'value' => floatval(key($params['mark'])), 'tip' => current($params['mark']).' (#val#)')));
			$ofc->add_element($mark);
		}
		
		$json = $ofc->toPrettyString();
		$chartDataIndex = $this->numCharts-1;
		// echo "<pre>$json</pre>"; // DEBUG
		$chart .= $this->Html->scriptBlock(
"swfobject.embedSWF('".$this->Html->url('/flash/open-flash-chart.swf')."', 'ofc-delayed-{$this->numCharts}', '100%', '100%', '9.0.0', '', { 'get-data':'ofc_get_data', 'id':'ofc_chart_{$this->numCharts}' });
chartData.ofc_chart_{$this->numCharts} = $json;\n");
		
		return "<div class=\"chart\" style=\"width: {$width}px; height: {$height}px;\">$chart</div>";
	}
	
	private function addDataToChart(&$ofc, $data, $params) {
		// Actual data
		foreach ($data as $set) {
			if ($params['type'] == 'bar') {
				$bar = new OFC_Charts_Bar();
				$bar->set_colour('#'.self::DATA_COLOR);
				$bar->set_values($set);
				// The following is quite a hack since the PHP5 library isn't
				// supposed to support animations, but json_encode wraps this
				// rather nicely
				$bar->{'on-show'} = array('type' => 'grow-up', 'cascade' => 1.2, 'delay' => 0.5);
				$ofc->add_element($bar);
			} else if ($params['type'] == 'line') {
				$bar = new OFC_Charts_Line();
				$bar->set_colour('#'.self::DATA_COLOR);
				$bar->{'dot-style'} = array ('type' => 'hollow-dot', 'dot-size' => 3, 'halo-size' => 2);
				$bar->set_values($set);
				// The following is quite a hack since the PHP5 library isn't
				// supposed to support animations, but json_encode wraps this
				// rather nicely
				$bar->{'on-show'} = array('type' => 'grow-up', 'cascade' => 1.2, 'delay' => 0.5);
				$ofc->add_element($bar);
			}
		}
	}
	
	private function getOrientation($axis, $or) {
		if ($or == 'V'
			|| ($axis == 'x' && $or == 'T')
			|| ($axis == 'y' && $or == '//'))
			return 'vertical';
		else
			return 'horizontal';
	}
	
	public function afterRender() {
		if ($this->numCharts > 0)
			echo $this->Html->scriptBlock(
				"var TOTAL_NUM_CHARTS = {$this->numCharts};"
			);
	}
}

?>