/* Copyright 2007-2012 Xavier 'Xr' Dalem
 * 
 * This file is part of Demyo. Demyo is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 * 
 * Demyo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU GPLv3 for more details.
 */
/*
jquery.advancedsearch

A generic interface to power searches based on criterias of various types.
 */
(function($) {
	$.fn.extend({
		advancedSearch: function(settings) {
			return this.each(setUpField);

			function setUpField() {
				$(this).submit(serializeFullCriteria);
				$(this).prepend(getGroup('or', true)); // Start with an "OR" clause, that's what most people want
			}

			/**
			 * Returns a group of fields (logical and/or), complete with selector for more sub-fields.
			 */
			function getGroup(type, first) {
				if (type == 'and') {
					typeString = settings.locale.all;
				} else if (type == 'or') {
					typeString = settings.locale.any;
				}

				// Create fieldset and type
				var fieldset = $('<fieldset></fieldset>');
				fieldset.attr('data-groupType', type);
				fieldset.append('<legend>' + typeString + '</legend>');
				var fieldSelection = $('<div class="fieldSelection"></div>');

				// Create field selector
				var select = $('<select></select>');
				$.each(settings.searchFields, function() {
					select.append('<option value="' + this.field + '">' + this.fieldName + '</value>');
				});
				// Two special options
				select.append('<option value="or">' + settings.locale.any + '</value>');
				select.append('<option value="and">' + settings.locale.all + '</value>');
				fieldSelection.append(select);

				// Create adding button
				var adder = $('<span class="adder">' + settings.locale.add + '</span>');
				adder.click(function() {
					addCriteria(fieldSelection, select.val());
				});
				fieldSelection.append(adder);
				fieldset.append(fieldSelection);

				// Create group remover
				if (!first) {
					var groupRemover = $('<span class="groupRemover">' + settings.locale.remove + '</span>');
					groupRemover.click(function() {
						$(fieldset).remove();
					});
					fieldset.append(groupRemover);
				}

				return fieldset;
			}

			/**
			 * Adds a criteria line based on the given field selector value. The line features a button to remove itself.
			 */
			function addCriteria(fieldSelection, field) {
				if (field == 'or' || field == 'and') {
					var group = getGroup(field, false);
					fieldSelection.before(group);
					return;
				}

				var div = $('<div class="criteria"></div>');
				div.attr('data-field', field);

				var searchField = getSearchField(field);
				var criteriaLine;
				switch (searchField.type) {
					case 'string':
						criteriaLine = appendStringCriteria(div, searchField);
						break;
					case 'boolean':
						criteriaLine = appendBooleanCriteria(div, searchField);
						break;
					case 'nullable':
						criteriaLine = appendNullableCriteria(div, searchField);
						break;
					case 'date':
						criteriaLine = appendDateCriteria(div, searchField);
						break;
					case 'integer':
						criteriaLine = appendIntegerCriteria(div, searchField);
						break;
					default:
						criteriaLine = null;
				}

				// Create remove button
				var remover = $('<span class="remover">' + settings.locale.remove + '</span>');
				remover.click(function() {
					div.remove();
				});
				div.append(remover);

				fieldSelection.before(div);
			}

			/**
			 * Adds a criteria for a string value.
			 */ 
			function appendStringCriteria(div, searchField) {
				appendFieldName(div, searchField);
				div.append(createOperator([
					{ value: 'LIKE', text: settings.locale.type_string_contains },
					{ value: 'NOT LIKE', text: settings.locale.type_string_not_contains },
					{ value: '=', text: settings.locale.type_string_is }
				]));
				div.append('<input type="text" class="value string" />');
			}

			/**
			 * Adds a criteria for a boolean value.
			 */
			function appendBooleanCriteria(div, searchField) {
				appendFieldName(div, searchField);
				div.append(createOperator([
					{ value: '=', text: settings.locale.type_boolean_true },
					{ value: '<>', text: settings.locale.type_boolean_false }
				]));
				div.append('<input type="hidden" class="value" value="true" />');
			}
			
			/**
			 * Adds a criteria for a nullable value.
			 */
			function appendNullableCriteria(div, searchField) {
				appendFieldName(div, searchField);
				div.append(createOperator([
					{ value: 'IS NULL', text: settings.locale.type_nullable_null },
					{ value: 'IS NOT NULL', text: settings.locale.type_nullable_not_null }
				]));
				div.append('<input type="hidden" class="value" value="" />');
			}

			/**
			 * Adds a criteria for a date value.
			 */
			function appendDateCriteria(div, searchField) {
				appendFieldName(div, searchField);
				div.append(createOperator([
					{ value: '>=', text: settings.locale.type_date_geq },
					{ value: '<=', text: settings.locale.type_date_leq }
				]));
				var input = $('<input type="text" class="value date" />');
				div.append(input);
				jQuery(input).DemyoDatePicker({ locale: settings.locale.type_date_locale });
			}
			
			/**
			 * Adds a criteria for an integer value.
			 */
			function appendIntegerCriteria(div, searchField) {
				appendFieldName(div, searchField);
				div.append(createOperator([
					{ value: '=', text: settings.locale.type_number_eq },
					{ value: '>', text: settings.locale.type_number_gt },
					{ value: '>=', text: settings.locale.type_number_geq },
					{ value: '<', text: settings.locale.type_number_lt },
					{ value: '<=', text: settings.locale.type_number_leq }
				]));
				var input = $('<input type="text" class="value number" />');
				// Set an ID: numberField requires it
				jQuery(input).attr('id', 'adv_search_' + Math.random().toString().replace(/\./, '_'));
				div.append(input);
				jQuery(input).numberField();
			}

			/**
			 * Adds the human-readable name of the field. 
			 */
			function appendFieldName(div, searchField) {
				div.append('<span class="fieldName">' + searchField.fieldName + '</field>');
			}

			/**
			 * Creates a select element to choose the comparison operator for a field.
			 */
			function createOperator(options) {
				var operator = $('<select class="operator"></select>');
				$.each(options, function() {
					operator.append('<option value="' + this.value + '">' + this.text + '</value>');
				});
				return operator;
			}

			/**
			 * Returns the search field options based on the field name.
			 */
			function getSearchField(field) {
				for (var i = 0; i < settings.searchFields.length; i++) {
					if (settings.searchFields[i].field == field) {
						return settings.searchFields[i];
					}
				}
				return null;
			}
			
			function serializeFullCriteria() {
				var serialized = serializeGroup($('> fieldset[data-groupType]', this));
				var jsonSerialized = JSON.stringify(serialized);
				// alert(jsonSerialized);
				if (serialized == null) {
					return false;
				}
				var criteriaInput = $('<input type="hidden" name="jsonCriteria" />');
				criteriaInput.val(jsonSerialized);
				$(this).append(criteriaInput);
				return true;
			}
			
			function serializeGroup(group) {
				var serialized = {};
				
				var components = [];
				
				$('> div.criteria', group).each(function() {
					var criteria = serializeSingleCriteria(this);
					if (criteria != null) {
						components.push(criteria);
					}
				});
				
				$('> fieldset[data-groupType]', group).each(function() {
					var subGroup = serializeGroup(this);
					if (subGroup != null) {
						components.push(subGroup);
					}
				});
				
				if (!components.length) {
					// Skip empty groups
					return null;
				}
				
				serialized[$(group).attr('data-groupType')] = components;
				
				return serialized;
			}
			
			function serializeSingleCriteria(criteria) {
				return {
					field: $(criteria).attr('data-field'),
					operator: $('select.operator', criteria).val(),
					value: $('input.value', criteria).val()
				};
			}
		}
	})
})(jQuery);

/* Locales. */
jQuery.extend({
	AdvancedSearchLocales: {
		eng: {
			any: "Matches any",
			all: "Matches all",
			add: "Add",
			remove: "Remove",
			type_string_is: "is",
			type_string_contains: "contains",
			type_string_not_contains: "does not contain",
			type_boolean_true: "yes",
			type_boolean_false: "no",
			type_date_geq: "later than",
			type_date_leq: "earlier than",
			type_date_locale: jQuery.DatePickerLocales.eng,
			type_number_eq: "equal to",
			type_number_gt: "greater than",
			type_number_geq: "greater or equal to",
			type_number_lt: "lower than",
			type_number_leq: "lower or equal to",
			type_nullable_null: "without",
			type_nullable_not_null: "with"
		},
		
		fre: {
			any: "Correspond à au moins un",
			all: "Correspond à tous",
			add: "Ajouter",
			remove: "Supprimer",
			type_string_is: "est",
			type_string_contains: "contient",
			type_string_not_contains: "ne contient pas",
			type_boolean_true: "oui",
			type_boolean_false: "non",
			type_date_geq: "plus tard que",
			type_date_leq: "plus tôt que",
			type_date_locale: jQuery.DatePickerLocales.fre,
			type_number_eq: "égal à",
			type_number_gt: "supérieur à",
			type_number_geq: "supérieur ou égal à",
			type_number_lt: "inférieur à",
			type_number_leq: "inférieur ou égal à",
			type_nullable_null: "sans",
			type_nullable_not_null: "avec"
		}
	}
});